{ ******************************************************************* }
{ Turbo Pascal Version 6.0 -> 7.0  AVL Tree Objects and Methods       }
{ ------------------------------------------------------------------- }
{ Copyright 1992, 1996 Roger E. Donais        <rdonais@southeast.net> }
{ ------------------------------------------------------------------- }

UNIT AVL;
{ ------------------------------------------------------------------- }
{$X+}                             { This unit requies extended syntax }
{ ------------------------------------------------------------------- }
{ A number of private methods and variables in the objects contained  }
{ in this unit are made public because the compiler does not support  }
{ a Friend type method.  A number of private structures are declared  }
{ in the interface to enable appropropriate type checking and support }
{ friends and descendent objects.                             ...red  }
{ ------------------------------------------------------------------- }
INTERFACE
USES OBJECTS, UTIL;

CONST  { tAsort status }
       asDONE = 0; { stOK }
       asGET  = 1;
       asPUT  = 2;

       { tAsort options }

       asDUPS       = 1;      { Allow duplicates                      }
       asLowSearch  = 2;      { End search on left-most duplicate     }
       asHighSearch = 4;      { End search on right-most duplicate    }
       asAnySearch  = 6;      { End search w/ first match encountered }

{ When duplicates are allowed, they are alway inserted to the right.  }
{ To disable duplicates, AND options with NOT(asDUPS).  By default,   }
{ Search returns the first matching node that it encounters.          }
{ To return the lowest matching node (oldest or first inserted),      }
{ OR options with asLowSearch, and to have Search position to the     }
{ highest (newest or most recently inserted) matching node,           }
{ OR options with asHighSearch.  To remove previously enabled Search  }
{ position options, logically AND options with NOT(asAnySearch).      }


{$IFDEF AVL_DEBUG}
       Adjust : String = ' 0 0 0 0 0';                  { Flags  1st & 2nd rotations  -2 -1  0 +1 +2 }
       Tally  : Array[-2..+2] Of Integer = (0,0,0,0,0); { Counts Number of rotations  -2 -1  0 +1 +2 }
       Rotate : String = '';                            { Flags type of last rotation -2 -1 ** +1 +2 }
{$ENDIF}

TYPE
      tpAVLnode = ^tAVLnode;
      tAVLnode  = RECORD
            Lft : tpAVLnode;
            Rgt : tpAVLnode;
            Case Integer Of
              1: (xxx : Array[0..2] of Byte; { reserved for index }
                  Bal : ShortInt);           { Node balance       }
              2: (Idx : Longint;             { 24-Bit Index       }
                  Itm : Pointer;             { Pointer to Data    }
                  Nxt : tpAVLnode);
      End;

      tpStrategy = ^tStrategy;
      tStrategy  = OBJECT( tObject )
           { -------------------------------------------------------- }
           { Maintains Pointer to an pObject/Record appended to Node  }
           { -------------------------------------------------------- }
           Function  KeyOf(TheItem: Pointer): Pointer;          VIRTUAL;
             { Given an Item, KeyOf should return TheItem's key.  The }
             { default KeyOf simply returns TheItem.                  }

           Function  Compare(Key1, Key2: Pointer): Integer;     VIRTUAL;
             { ABSTRACT: Descendants obj's should return a result of: }
             { -1: Key1 < Key2,  0: Key1 = Key2,  or  +1: Key1 > Key2 }
             { Key1 and Key2 are pointers returned by KeyOf method    }
             { Make certain that Compare returns all possible values. }
             { Even a tree that will never hold duplicate items needs }
             { to return 0 or Search will not function properly.      }

           Function  GetItem(VAR S: tStream): Pointer;          VIRTUAL;
             { Given a Stream, GetItem should read an Item and return }
             { a pointer to it.  Defailt GetItem uses Stream.Get      }

           Procedure PutItem(VAR S: tStream; TheItem: Pointer); VIRTUAL;
             { Given a Stream and Item, PutItem should write TheItem  }
             { onto stream. Default PutItem uses Stream.Put(TheItem)  }

           Procedure FreeItem(TheItem: Pointer);                VIRTUAL;
             { Given an Item, FreeItem should dispose TheItem.        }
             { Default FreeItem invokes pObject(TheItem)^.Free        }
      End;

      tpCollectionStrategy = ^tCollectionStrategy;
      tCollectionStrategy  = OBJECT( tStrategy )
           { --------------------------------------------------------------- }
           { Maintains Pointer to an Item (tObject/Record) appended to Node  }
           { using methods imported from a tSortedCollection                 }
           { --------------------------------------------------------------- }
           Constructor Init(Dummy: pSortedCollection);
           Destructor  Done;                                    VIRTUAL;

           Function  KeyOf(TheItem: Pointer): Pointer;          VIRTUAL;
           Function  Compare(Key1, Key2: Pointer): Integer;     VIRTUAL;
           Function  GetItem(VAR S: tStream): Pointer;          VIRTUAL;
           Procedure PutItem(VAR S: tStream; TheItem: Pointer); VIRTUAL;
           Procedure FreeItem(TheItem: Pointer);                VIRTUAL;
           PRIVATE
             User: pSortedCollection;
      End;

      tpAVLCollection = ^tAVLCollection;
      tAVLCollection = OBJECT( tObject )
           { -------------------------------------------------------- }
           { Maintains basic AVL tree supporting indexed access ...   }
           { -------------------------------------------------------- }
           Item    : Pointer;  { RO: PointerOf last accessed Item     }
           Index   : Longint;  { RO: IndexOf last accessed Item       }
           Count   : Longint;  { RO: Number of nodes in the tree      }
           Options : Word;     { RW: Default = asDUPS                 }

           Constructor Init(Strategy: tpStrategy);
           Destructor  Done;                                    VIRTUAL;

           Procedure FreeAll;
             { All entries are removed from the tree and strategy's   }
             { FreeItem is called to dispose each Item.  Item is set  }
             { to NIL and Count is set to 0.                          }

           Procedure DeleteAll;
             { All entries are removed from the tree.  Item is set to }
             { NIL and Count is set to 0.                             }

           Function  Insert(TheItem: Pointer): Boolean;
             { Use search strategy to locate TheItem's position w/in  }
             { the tree.  If an entry exists and duplicates are not   }
             { allowed, Insert calls strategy's FreeItem to dispose   }
             { of the item and returns FALSE.  Otherwise TRUE is      }
             { returned after constructing and inserting a new node   }
             { w/ TheItem into the tree, and incrementing Count.      }

           Procedure AtInsert(Subscript: Longint; TheItem: Pointer);
             { Constructs a new node w/ TheItem and inserts it at the }
             { specified position.  Subscripts <= 1 cause the node to }
             { be added as the new left-most node and values larger   }
             { than the actual number of nodes in the tree casue the  }
             { node to be added as the new right-most node.  Count is }
             { incremented, but is not used to determine where the    }
             { mode should be inserted.                               }

           Procedure AtFree(Subscript: Longint);
             { Returns TRUE w/ Item set to NIL and index set to 0 if  }
             { specified node number is found and the node is removed }
             { from the tree, strategy's FreeItem is called to        }
             { dispose the Item, and Count is decremented.  Count is  }
             { *NOT* used to locate the node w/in the tree.  If the   }
             { specified node does not exist w/in the tree, AtFree    }
             { returns FALSE and Item and Index are unchanged.        }

           Procedure AtDelete(Subscript: Longint);
             { Returns TRUE w/ Item and index set, if the specified   }
             { node number is found and the node is removed from the  }
             { tree and Count is decremented.  Count is *NOT* used to }
             { locate the node w/in the tree.  If the specified node  }
             { does not exist w/in the tree, AtDelete returns FALSE   }
             { and Item and Index are unchanged.                      }

           Function  AtItem(Subscript: Longint): Boolean;
             { Returns TRUE w/ Item and index set if the specified    }
             { node number is found. Count is *NOT* used to locate    }
             { the node w/in the tree. If the specified node does not }
             { exist w/in the tree, AtItem returns FALSE and Item and }
             { Index are unchanged.                                   }

           Function  IndexOf(TheItem: Pointer): Boolean;
             { An attempt is made to locate an Item pointer matching  }
             { TheItem using the normal Search strategy.  If this     }
             { fails, each node is checked for a matching pointer.    }
             { Returns TRUE w/ Item and index set if a matching       }
             { pointer is found, otherwise returns FALSE w/ Item      }
             { set to NIL and Index set to zero.                      }

           Function  Search(ItemKey: Pointer): Boolean;
             { Returns TRUE w/ Item and index set if ItemKey is found }
             { otherwise returns FALSE w/Item and Index undefined.    }
             { Found position depends upon asXXXX search options      }

           Procedure ForEach(Action: Pointer);
             { (Sub)Procedure Action(TheItem: Pointer); FAR;          }
             { Entire tree is traversed in order of LEFT, RIGHT, NODE }
             { Item and Index are *NOT* valid on any call to Action   }
             { and are not changed by this method                     }

           Procedure Traverse(Fwd: Boolean; Action: Pointer);
             { (Sub)Procedure Action(Node: tpAVLnode; Depth: Word); FAR; }
             { Tree is traversed in order: from LEFT, NODE, RIGHT     }
             { if Fwd is TRUE, and from RIGHT, NODE, LEFT when FALSE. }
             { Initially Item is set to NIL and Index is set to zero. }
             { Item and Index are reset before each call to action.   }

           Function  FirstThat(Action: Pointer): Boolean;
             { (Sub)Function Action(TheItem: Pointer); Boolean; FAR;  }
             { Tree is traversed starting with the left-most node     }
             { in order from LEFT, NODE, RIGHT until action is TRUE   }
             { or the entire tree has been processed                  }
             { Initially Item is set to nil and Index is set to Zero. }
             { Item and Index are set before each call to action      }

           Function  LastThat(Action: Pointer): Boolean;
             { (Sub)Function Action(TheItem: Pointer); Boolean; FAR;  }
             { Tree is traversed starting with the right-most node    }
             { in order from RIGHT, NODE, LEFT until action is TRUE   }
             { or the entire tree has been processed                  }
             { Initially Item is set to nil and Index is set to Zero. }
             { Item and Index are set before each call to action      }

           Function  NextThat(Subscript: Longint; Action: Pointer): Boolean;
             { (Sub)Function Action(TheItem: Pointer); Boolean; FAR;  }
             { tree is traversed starting with next higher Subscript  }
             { in order from LEFT, NODE, RIGHT until action is TRUE   }
             { or the entire tree has been processed                  }
             { Initially Item is set to nil and Index is set to Zero. }
             { Item and Index are set before each call to action      }

           Function  PrevThat(Subscript: Longint; Action: Pointer): Boolean;
             { (Sub)Function Action(TheItem: Pointer); Boolean; FAR;  }
             { tree is traversed starting with next lower Subscript   }
             { in order from RIGHT, NODE, LEFT until action is TRUE   }
             { or the entire tree has been processed                  }
             { Initially Item is set to nil and Index is set to Zero. }
             { Item and Index are set before each call to action      }

           PRIVATE
              Root      : tpAVLnode;  { RO: Root node of tree         }
          {!} Link      : tpAVLnode;  { RO: Local parameter           }
              User      : tpStrategy; { RO: Mantenaince strategy      }

              Procedure BuildNode(VAR Node: tpAVLnode; TheItem: Pointer);
      End;

      tpBlock = ^tBlock;
      tBLOCK  = RECORD
          { Private structure used by tAsort }
          Count : Longint;
          Offset: Longint;
          Item  : Pointer;
      End;

      tpAsort = ^tASort;
      tAsort  = OBJECT( tObject )
           MemL  : Longint;         { Low memory limit...             }
           Sort  : tpAVLCollection; { AVL tree & Current Strategy     }

           Constructor Init(Strategy: tpStrategy; S: pStream; MemLimit: Longint);
           Destructor  Done;                                    VIRTUAL;

           Procedure Put(TheItem: Pointer);
             { Given an item, Put adds the item to the current sort   }

           Function  Get(VAR TheItem: Pointer): Boolean;
             { Returns FALSE if there are no more items, or if a      }
             { stream error occured, otherwise places the next sorted }
             { item into TheItem and returns TRUE.                    }

           Function  Status: Integer;
             { Returns the current stXXXX constant for the Stream, or }
             { if stOK, returns the appropriate asXXXX state constant.}
             { When tAsort has successfully initialized, Status is    }
             { asPUT indicating that it is in "PUT" mode, and will    }
             { accept items to be sorted. State will remain asPUT     }
             { until a stream error occurs or after the first call to }
             { Get.  After each successful call to Get, state will    }
             { return asGET.  Otherwise state will report an stXXXX   }
             { constant if there was a stream error, or asDONE when   }
             { *NO* additional items remain.                          }

           Function  OnStream: Boolean;
             { Returns FALSE if entire sort results are in memory and }
             { returns TRUE if portions were written to the stream.   }

           Function  MergeBlocks: Word;
             { Returns the number of merge blocks that are currently  }
             { on the stream.                                         }

           Procedure Flush;
             { Writes current tree to the Stream.                     }

           Procedure FreeItem(TheItem: Pointer);
             { Invokes the user supplied strategy to dispose of an    }
             { Item.  FreeItem should be called to dispose of TheItem }
             { returned by Get, when TheItem is no longer required.   }

           PRIVATE
             NBlocks : Word;            { Number of Merge Blocks      }
               State : Integer;         { Current state               }
                User : tpStrategy;      { User Strategy               }
                TmpF : pStream;         { User supplied Stream        }
                List : Pointer;         { List of blocks to be merged }

             Procedure Merge(Block: tpBlock);
      End;

{ ******************************************************************* }
{ Initialization requires the user to supply a pStream to contain the }
{ intermediate results of a large sort, and a tpStrategy that will    }
{ determine the sort order, and specify the minimum amounnt of memory }
{ to be reserved.                                                     }
{                                                                     }
{ The tStrategy must have an appropriate KeyOf, Compare, PutItem,     }
{ GetItem, and FreeItem methods.                                      }
{                                                                     }
{ tAsort assumes that all items being sorted have been dynamically    }
{ allocated on the heap.  During a large sort it will invoke the      }
{ Straategy's PutItem to have items written to the stream, will       }
{ dispose the item by invoking Strategy's FreeItem method, and will   }
{ invoke Strategy's GetItem method to read items from the stream.     }
{                                                                     }
{ Status returns the current stXXXX constant for the Stream, or if    }
{ stOK, returns the appropriate asXXXX state constant.  When tAsort   }
{ has successfully initialized, it's Status returns asPUT indicating  }
{ that it is in "PUT" mode, and will accept items to be sorted. State }
{ will remain asPUT until a stream error occurs or after the first    }
{ call to Get.  After each successful call to Get, state will return  }
{ asGET.  Otherwise state will report an stXXXX constant if there was }
{ a stream error, or asDONE when *NO* additional items remain.        }
{                                                                     }
{ When Done, tAsort will dispose of both the Strategy and Stream,     }
{ Note, however, that if the stream involved a DOS File that          }
{ the file will not have been deleted unless it was done by the       }
{ stream's destructor.                                                }
{ ******************************************************************* }


{ STRATEGY METHODS }
CONST   Done_Method       = 0 * Sizeof(Pointer) + VMTheaderSize;
        KeyOf_Method      = 1 * Sizeof(Pointer) + VMTheaderSize;
        Compare_Method    = 2 * Sizeof(Pointer) + VMTheaderSize;
        GetItem_Method    = 3 * Sizeof(Pointer) + VMTheaderSize;
        PutItem_Method    = 4 * Sizeof(Pointer) + VMTheaderSize;
        FreeItem_Method   = 5 * Sizeof(Pointer) + VMTheaderSize;

{ As seen from the above the implementation of these objects depends  }
{ upon the order that virtual methods for tObject and tCollection     }
{ where declared.  The above is the original order established by     }
{ Borland.  If you have altered this order, you must make the         }
{ appropriate adjustments to the above declarations.          ...red  }
{ ------------------------------------------------------------------- }

{ ******************************************************************* }
IMPLEMENTATION


{ =================================================================== }
{                   S T R A T E G Y    M E T H O D S                  }
{ =================================================================== }

FUNCTION tStrategy.KeyOf(TheItem: Pointer): Pointer;
{ =================================================================== }
BEGIN KeyOf := TheItem;
END;


FUNCTION tStrategy.Compare(Key1, Key2: Pointer): Integer;
{ =================================================================== }
BEGIN abstract
END;


FUNCTION tStrategy.GetItem(VAR S: tStream): Pointer;
{ =================================================================== }
BEGIN GetItem := S.Get;
END;


PROCEDURE tStrategy.PutItem(VAR S: tStream; TheItem: Pointer);
{ =================================================================== }
BEGIN S.Put(TheItem);
END;


PROCEDURE  tStrategy.FreeItem(TheItem: Pointer);
{ =================================================================== }
BEGIN pObject(TheItem)^.Free;
END;


{ =================================================================== }
{              C O L L E C T I O N     S T R A T E G Y                }
{ =================================================================== }

CONSTRUCTOR tCollectionStrategy.Init(Dummy: pSortedCollection);
{ =================================================================== }
BEGIN
    inherited Init;
    User := Dummy;
END;


DESTRUCTOR tCollectionStrategy.Done;
{ =================================================================== }
BEGIN
    Dispose(User, Done);
    inherited Done;
END;


FUNCTION tCollectionStrategy.KeyOf(TheItem: Pointer): Pointer;
{ =================================================================== }
BEGIN KeyOf := User^.KeyOf( TheItem );
END;


FUNCTION tCollectionStrategy.Compare(Key1, Key2: Pointer): Integer;
{ =================================================================== }
BEGIN Compare := User^.Compare(Key1, Key2);
END;


FUNCTION tCollectionStrategy.GetItem(VAR S: tStream): Pointer;
{ =================================================================== }
BEGIN GetItem := User^.GetItem(S);
END;


PROCEDURE tCollectionStrategy. PutItem(VAR S: tStream; TheItem: Pointer);
{ =================================================================== }
BEGIN User^.PutItem(S, TheItem);
END;


PROCEDURE  tCollectionStrategy.FreeItem(TheItem: Pointer);
{ =================================================================== }
BEGIN User^.FreeItem(TheItem);
END;


{ =================================================================== }
{         L O C A L     S U P P O R T     F U N C T I O N S           }
{ =================================================================== }

{\FUNCTION SignOf(i: Integer): Integer; Assembler; {}
{\{ =================================================================== } {}
{\ASM {}
      {\mov    AX, 1 {}
      {\shl    [ i ], 1 {}
      {\ja     @@1                { Above means C=0, Z=0, thus AX=1 } {}
      {\sbb    AX, AX             { Below or Equal makes AX=-1 or 0 } {}
{\@@1: {}
{\END; {}


PROCEDURE tAVLcollection.BuildNode(VAR Node: tpAVLnode; TheItem: Pointer);
{ =================================================================== }
BEGIN
    New(Node);
    FillChar(Node^, Sizeof(Node^), 0);
    Node^.Itm := TheItem;
    {!}
    If Ofs(Link^) and 7 = 0 Then Begin
       Node^.Nxt := Link;
       Link := Node;
    End;
    {}
END;

{ =================================================================== }
{               A V L     S O R T     S T R A T E G Y                 }
{ =================================================================== }
TYPE   tpAsortStrategy = ^tAsortStrategy;
       tAsortStrategy  = OBJECT( tStrategy )

           Constructor Init(TheSort: tpAVLCollection);
           Function    Compare(Key1, Key2: Pointer): Integer;  VIRTUAL;
           Procedure   FreeItem(TheItem: Pointer);             VIRTUAL;
           PRIVATE
              User : tpStrategy;                 { Taken from TheSort }
              Sort : tpAVLCollection;            { Asort's AVL tree   }
       End;


CONSTRUCTOR tAsortStrategy.Init(TheSort: tpAVLCollection);
{ =================================================================== }
BEGIN
    inherited Init;
    User := TheSort^.User;
    Sort := TheSort;
END;


FUNCTION tAsortStrategy.Compare(Key1, Key2: Pointer): Integer;
{ =================================================================== }
VAR i: Integer;
BEGIN
    i := Signof(User^.Compare(User^.KeyOf(tpBlock(Key1)^.Item), User^.KeyOf(tpBlock(Key2)^.Item)));
    If (i = 0) and (Sort^.Options and asDUPS <> 0) Then
       i := Signof(tpBlock(Key1)^.Offset - tpBlock(Key2)^.Offset);
    Compare := i;
END;


PROCEDURE tAsortStrategy.FreeItem(TheItem: Pointer);
{ =================================================================== }
{ tAsortStrategy.FreeItem is called when the tAVLcollection.Insert    }
{ needs to discard a duplicate entry being merged.  The user item is  }
{ to be disposed, but the Block record is to be retained.             }
{ =================================================================== }
BEGIN
    If tpBlock(TheItem)^.Item <> NIL Then
       User^.FreeItem(tpBlock(TheItem)^.Item);
END;


{ =================================================================== }
{                 A V L     S O R T     M E T H O D S                 }
{ =================================================================== }

CONSTRUCTOR tAsort.Init(Strategy: tpStrategy; S: pStream; MemLimit: Longint);
{ =================================================================== }
BEGIN
    inherited Init;
    {S^.Write(MemL, 1024);}
    S^.Seek( 0 );
    TmpF  := S;
    MemL  := MemLimit;
    State := asPUT;
    User  := Strategy;
    Sort  := New(tpAVLCollection, Init(Strategy));
END;


DESTRUCTOR  tAsort.Done;
{ =================================================================== }

    PROCEDURE DisposeBlock(TheBlock: tpBlock); FAR;
    { --------------------------------------------------------- }
    BEGIN
        FreeItem(TheBlock^.Item);     { Dispose User Item   }
        Dispose(TheBlock);            { Dispose Merge Block }
    END;

    PROCEDURE DisposeList(VAR List: tpBlock);
    { --------------------------------------------------------- }
    VAR p: tpBlock;
    BEGIN
        While List <> NIL Do Begin
           p := List; List := p^.Item; Dispose(p);
        End;
    END;

VAR p: tpBlock;
BEGIN
    If TypeOf(Sort^.User^) = TypeOf(tAsortStrategy)  Then Begin
       Sort^.ForEach(@DisposeBlock);
       Sort^.Free;
    End Else
       Sort^.FreeAll;
    DisposeList(tpBlock(List));
    TmpF^.Free;
    inherited Done;
END;


PROCEDURE tAsort.Flush;
{ =================================================================== }
VAR p: tBlock;

    FUNCTION FlushNode(TheItem: Pointer): Boolean; FAR;
    { --------------------------------------------------------- }
    BEGIN
        User^.PutItem(TmpF^, TheItem);
        FlushNode := (TmpF^.Status <> 0);
    END;

BEGIN
    If (Status = asPUT) and (Sort^.Count > 0) Then Begin
       p.Count  := Sort^.Count;      { Insert new block at head of list }
       p.Offset := TmpF^.GetPos;
       p.Item   := List;
       Sort^.FirstThat(@FlushNode);
       Sort^.FreeAll;
       New(tpBlock(List));
       tpBlock(List)^ := p;
       Inc(NBlocks);
       Word(State) := Word(State) or $8000;
    End;
END;


PROCEDURE tAsort.Merge(Block: tpBlock);
{ =================================================================== }
{ Note the AVLCollection strategy has been replaced with an Asort     }
{ Strategy.  As such, the merge block will be retained even though    }
{ duplicate user items may be disposed. The retention of the merge    }
{ block reqires it be disposed by this method when it has expired.    }
{ =================================================================== }
BEGIN
    If (Block^.Count > 0) Then Begin
       TmpF^.Seek(Block^.Offset);
       While (Block^.Count > 0) Do Begin
          Dec(Block^.Count);
          Block^.Item := {!tpAsortStrategy(Sort^.User)^.{}User^.GetItem(TmpF^);
          If Sort^.Insert(Block) Then Begin
             Block^.Offset := TmpF^.GetPos;
             Exit;
          End;
       End;
    End;
    Dispose(Block);
END;


PROCEDURE tAsort.Put(TheItem: Pointer);
{ =================================================================== }
BEGIN
    If Status = asPUT Then Begin
       If (MemL > 0) and (MaxAvail < MemL) Then Flush;
       Sort^.Insert(TheItem)
    End;
END;


FUNCTION tAsort.Get(VAR TheItem: Pointer): Boolean;
{ =================================================================== }
LABEL  AGAIN;
VAR Block: tpBlock;

BEGIN
    Get := FALSE;                           { Assume failure          }
    If Lo(State) = asPUT Then
    Case Hi(State) of
       $80: BEGIN
                Flush;                      { Flush tree to stream    }
                TmpF^.Flush;                { Flush stream (to disk?) }
                Dec(State);                 { Change state to asGet   }
                Sort^.User := New(tpAsortStrategy, Init(Sort));
                If TmpF^.Status < stOK Then Exit;
                While List <> NIL Do Begin
                   Block := List;
                   List := Block^.Item;
                   Merge(Block);
                End;
            END;
       $00: Dec(State);                     { Change state to asGet   }
    End{Case};
    Case Word(State) of
       $8001: BEGIN
                  { Merging intermediate blocks from stream }
                  Sort^.AtDelete(1);
                  If Sort^.Item <> NIL Then Begin
                     TheItem := tpBlock(Sort^.Item)^.Item;
                     Merge(Sort^.Item);
                     Get := TRUE;
                  End Else State := asDONE;
              END;
          $1: BEGIN
                  { Entire sort resides in memory }
                  Sort^.AtDelete(1);
                  If Sort^.Item <> NIL Then Begin
                     TheItem := Sort^.Item;
                     Get := TRUE;
                  End Else State := asDone;
              END;
    End{Case};
END;


FUNCTION tAsort.Status;
{ =================================================================== }
BEGIN
    If TmpF^.Status = stOK Then
       Status := Lo(State)
    Else Status := TmpF^.Status;
END;


FUNCTION tAsort.OnStream: Boolean;
{ =================================================================== }
BEGIN
    OnStream := State and $8000 <> 0;
END;


FUNCTION tAsort.MergeBlocks: Word;
{ =================================================================== }
BEGIN
    MergeBlocks := NBlocks;
END;


PROCEDURE tAsort.FreeItem(TheItem: Pointer);
{ =================================================================== }
{ This method is disposes of a user item regardless of the sort state }
{ =================================================================== }
BEGIN
    User^.FreeItem(TheItem);
END;

{ =================================================================== }
{           A V L     C O L L E C T I O N     M E T H O D S           }
{ =================================================================== }

CONSTRUCTOR tAVLCollection.Init(Strategy: tpStrategy);
{ =================================================================== }
BEGIN
    inherited Init;
    Options := asDUPS + asAnySearch;
    User := Strategy;
END;


DESTRUCTOR tAVLCollection.Done;
{ =================================================================== }
BEGIN
    FreeAll;
    Dispose(User, Done);
    inherited Done;
END;


{ =================================================================== }
{                B A S I C     A V L     M E T H O D S                }
{ =================================================================== }

PROCEDURE tAVLCollection.DeleteAll;
{ =================================================================== }

      PROCEDURE Discard(TheNode:  tpAVLnode); NEAR;
      { --------------------------------------------------------- }
      BEGIN Dispose(TheNode);
      END;

VAR p: tpAVLnode;
BEGIN
    {!} If Odd(Ofs(Link^)) Then {}
       ForEach(Ptr(0 ,Ofs(Discard)))
    {!} Else
       While Link <> NIL Do Begin
          p := Link;
          Link := p^.Nxt;
          Discard(p);
    End;
    Link := NIL;
    {}
    Root := NIL;
    Item := NIL;
    Count := 0;
    Index := 0;
END;


PROCEDURE tAVLCollection.FreeAll;
{ =================================================================== }

      PROCEDURE Discard(Node: tpAVLNode); NEAR;
    { --------------------------------------------------------- }
      BEGIN User^.FreeItem(Node^.Itm);
            Dispose(Node);
      END;

VAR p: tpAVLnode;
BEGIN

    {!} If Odd(Ofs(Link^)) Then {}
       ForEach(Ptr(0 ,Ofs(Discard)))
    {!} Else
       While Link <> NIL Do Begin
          p := Link;
          Link := p^.Nxt;
          Discard(p);
    End;
    Link := NIL;
    {}
    Root := NIL;
    Item := NIL;
    Count := 0;
    Index := 0;
END;


PROCEDURE tAVLCollection.AtFree(Subscript: Longint);
{ =================================================================== }
BEGIN
     AtDelete( Subscript );
     If Item <> NIL Then
        User^.FreeItem( Item );
     Item := NIL;
     Index := 0;
END;


PROCEDURE tAVLCollection.AtDelete(Subscript: Longint);
{ =================================================================== }
VAR Changed: Boolean;     { TRUE if tree depth has chaned, else FALSE }
    Deleted: Integer;     { 1 if deletion actually occured, else 0    }

      PROCEDURE DeletedRight(VAR p: tpAVLnode);
      { ------------------------------------------------------------- }
      VAR p1, p2: tpAVLnode;
          b1, b2: ShortInt;
      BEGIN
          { Changed is TRUE on ENTRY }
          Case p^.Bal Of
             +1: Begin p^.Bal := 0; End;
              0: Begin p^.Bal := -1; Changed := FALSE; END;
             -1: Begin
                    p1 := p^.Lft;
                    b1 := p1^.Bal;
                    If b1 <= 0 Then Begin { Single LL Rotation }
                       {$IFDEF AVL_DEBUG}
                       Rotate := Rotate + '+1LL';
                       {$ENDIF}
                       Dec(p^.Idx,p1^.Idx+1);
                       p^.Lft := p1^.Rgt;
                       p1^.Rgt := p;
                       If b1 = 0 Then Begin
                          p^.Bal := -1;
                          p1^.Bal := 1;
                          Changed := FALSE;
                       End Else Begin
                          p^.Bal := 0;
                          p1^.Bal := 0;
                       End;
                       p := p1;
                    End Else Begin { Double LR Rotation }
                       {$IFDEF AVL_DEBUG}
                       Rotate := Rotate + '+2LR';
                       {$ENDIF}
                       p2 := p1^.Rgt;
                       b2 := p2^.Bal;
                       Dec(p^.Idx,p1^.Idx+p2^.Idx+2);
                       Inc(p2^.Idx,p1^.Idx+1);
                       p1^.Rgt := p2^.Lft;
                       p2^.Lft := p1;
                       p^.Lft := p2^.Rgt;
                       p2^.Rgt := p;
                       If b2 = -1 Then p^.Bal := 1 Else p^.Bal := 0;
                       If b2 = +1 Then p1^.Bal := -1 Else p1^.Bal := 0;
                       p := p2;
                       p2^.Bal := 0;
                    End;
                End;
          End{Case};
      END{DeletedLeft};

      PROCEDURE DeletedLeft(VAR p: tpAVLnode);
      { ------------------------------------------------------------- }
      VAR p1, p2: tpAVLnode;
          b1, b2: ShortInt;
      BEGIN
          { Changed is TRUE on ENTRY }
          Case p^.Bal Of
             -1: Begin p^.Bal := 0; End;
              0: Begin p^.Bal := 1; Changed := FALSE; END;
             +1: Begin
                    p1 := p^.Rgt;
                    b1 := p1^.Bal;
                    If b1 >= 0 Then Begin { Single RR Rotation }
                       {$IFDEF AVL_DEBUG}
                       Rotate := Rotate + '-1RR';
                       {$ENDIF}
                       Inc(p1^.Idx, Succ(p^.Idx));
                       p^.Rgt := p1^.Lft;
                       p1^.Lft := p;
                       If b1 = 0 Then Begin
                          p^.Bal := 1;
                          p1^.Bal := -1;
                          Changed := FALSE;
                       End Else Begin
                          p^.Bal := 0;
                          p1^.Bal := 0;
                       End;
                       p := p1;
                    End Else Begin { Double RL Rotation }
                       {$IFDEF AVL_DEBUG}
                       Rotate := Rotate + '-2RL';
                       {$ENDIF}
                       p2 := p1^.Lft;
                       b2 := p2^.Bal;
                       Dec(p1^.Idx,Succ(p2^.Idx and $00FFFFFF));
                       Inc(p2^.Idx,Succ(p^.Idx and $00FFFFFF));
                       p1^.Lft := p2^.Rgt;
                       p2^.Rgt := p1;
                       p^.Rgt := p2^.Lft;
                       p2^.Lft := p;
                       If b2 = +1 Then p^.Bal := -1 Else p^.Bal := 0;
                       If b2 = -1 Then p1^.Bal := 1 Else p1^.Bal := 0;
                       p := p2;
                       p2^.Bal := 0;
                    End;
                End;
          End{Case};
      END{DeletedLeft};

      PROCEDURE Delete(VAR p: tpAVLnode);
      { ------------------------------------------------------------- }
      VAR q: tpAVLnode;
          k: Longint;

            PROCEDURE Del(VAR r: tpAVLnode);
            { ------------------------------------------------------- }
            BEGIN
                If r^.Rgt <> NIL Then Begin
                   Del(r^.Rgt);
                   If Changed Then DeletedRight(r);
                End Else Begin
                   Exchange(q^.Itm, r^.Itm, Sizeof(Pointer));
                   q := r;
                   r := r^.Lft;                 { Pull up replacedment }
                   Changed := TRUE;
                   Deleted := 1;
                End;
            END{Del};

        BEGIN{Delete}
            If p <> NIL Then Begin
               k := Succ(p^.Idx and $00FFFFFF);
               If k > Subscript Then Begin
                  Delete(p^.Lft);
                  Dec(p^.Idx, Deleted);
                  If Changed Then DeletedLeft(p);
               End Else
               If k < Subscript Then Begin
                  Dec(Subscript, k);
                  Delete(p^.Rgt);
                  If Changed Then DeletedRight(p);
               End Else Begin
                  q := p;
                  If q^.Rgt = NIL Then Begin
                     p := q^.Lft;
                     Changed := TRUE;
                     Deleted := 1;
                  End Else
                  If q^.Lft = NIL Then Begin
                     p := q^.Rgt;
                     Changed := TRUE;
                     Deleted := 1;
                  End Else Begin
                     Del(q^.Lft);
                     Dec(p^.Idx, Deleted);
                     If Changed Then DeletedLeft(p);
                  End;
                  Item := q^.Itm;  { Item from deleted node }
                  Dec(Count);
                  Dispose(q);
               End;
            End;
         END;

BEGIN{AtDelete}

    Changed := FALSE;
    Deleted := 0;
    Item := NIL;
    Delete(tpAVLnode(Root));
    {!} If Item <> NIL Then Begin
       Link := Ptr(0, 1);
       Index := Subscript;
    End Else
    {}
    Index := 0;

END{AtDelete};


FUNCTION tAVLCollection.IndexOf(TheItem: Pointer): Boolean;
{ =================================================================== }
VAR Hit: Longint;

      FUNCTION Matches(Item: Pointer): Boolean; FAR;
      { --------------------------------------------------------- }
      BEGIN Matches := (Item = TheItem);
      END;

BEGIN
    IndexOf := TRUE;   { Assume success }
    If Search(User^.KeyOf(TheItem)) Then Begin
       If (TheItem <> Item) Then Begin
          Hit := Index;
          If NOT NextThat(Hit, @Matches) Then Begin
             IndexOf := PrevThat(Hit, @Matches); { Index Zero on Failure }
          End; {Else assumption is correct}
       End; {Else assumption is correct}
    End Else
       IndexOf := FirstThat(@Matches);          { Index Zero on Failure }
END;


FUNCTION tAVLCollection.AtItem(Subscript: Longint): Boolean; Assembler;
{ =================================================================== }
ASM
        push  DS
        lds   SI, [ Self ]
        lea   SI, [ SI.Root ]
        xor   BX,BX
        xor   DX,DX
        jmp   @Left

@Right: mov   DX, AX
        mov   BX, CX
        lea   SI, [ tAVLNode(SI).Rgt ]

@Left:  lds   SI, [ tAVLNode(SI).Lft ]
        mov   CX, DS
        jcxz  @Done

        mov   CX, [ tAVLnode(SI).Idx.Word.0 ]
        mov   AL, [ tAVLnode(SI).Idx.Byte.2 ]
        xor   AH, AH
        stc
        adc   CX, BX
        adc   AX, DX

        cmp   [ Subscript.Word.2 ], AX
        jg    @Right
        jl    @Left

        cmp   [ Subscript.Word.0 ], CX
        ja    @Right
        jb    @Left

        { DS:SI is the indexed node and AX:CX conatins the index }
        les   DI, [ tAVLnode(SI).Itm ]
        lds   SI, [ Self ]
        mov   [ SI.Index.Word.0 ], CX
        mov   [ SI.Index.Word.2 ], AX
        mov   [ SI.Item.Word.0 ], DI
        mov   [ SI.Item.Word.2 ], ES
        mov   CX, 1

@Done:  mov   AX, CX
        pop   DS
END;


PROCEDURE tAVLCollection.ForEach(Action: Pointer); Assembler;
{ =================================================================== }
{ Action is usually defined as a FAR (Sub)Procedure that is called    }
{ w/ pointer to TheItem.  However, Action may be passed w/ SEG of 0   }
{ by methods in *THIS* unit to specify a local NEAR (Sub)Procedure    }
{ that is to be called /w a pointer to TheNode.                       }
{ ------------------------------------------------------------------- }
ASM
         les   DI, [ Self ]
         les   DI, [ ES:DI.Root ]
         mov   AX, OFFSET @Done
         push  AX

@Next:   mov   CX, ES
         jcxz  @Rtn

         push  ES                                { Save current node      }
         push  DI
         les   DI, [ ES:DI ]                     { And process left ...   }
         call  @Next

         pop   DI; pop  ES                       { Refresh node address   }
         push  ES; push DI
         les   DI, [ tAVLnode(ES:DI).Rgt ]       { And process right ...  }
         call  @Next

         mov   CX, [ Action.Word.2 ]
         jcxz  @Local                            { FreeNode if Action NIL }

         pop   DI
         pop   ES
         push  [ tAVLNode(ES:DI).Itm.Word.2 ]
         push  [ tAVLNode(ES:DI).Itm.Word.0 ]
         push  [ BP.Word.0 ]
         call  DWORD PTR [ Action ]
@Rtn:    retn

@Local:  push  [ BP.Word.0 ]
         call  WORD PTR [ Action ]
         retn
@Done:
END;


FUNCTION tAVLCollection.Search(ItemKey: Pointer): Boolean;
{ =================================================================== }
VAR j: Integer;
    p: tpAVLnode;
BEGIN
    p := tpAVLnode(Root);
    Item  := NIL;
    Index := 0;
    While p <> NIL Do Begin
       Item := p^.Itm;
       j := User^.Compare(ItemKey, User^.KeyOf(p^.Itm));
       If j < 0 Then
          p := p^.Lft
       Else Begin
          If j > 0 Then Begin
             Inc(Index, Succ(p^.Idx and $00FFFFFF));
             p := p^.Rgt
          End Else Begin
             Search := TRUE;
             Case Options and asAnySearch of
                {0, asAnySearch: nop}

                asLowSearch :
                   While (p^.Lft <> NIL)
                     and (User^.Compare(ItemKey, User^.KeyOf(p^.Lft^.Itm)) = 0) Do
                       p := p^.Lft;

                asHighSearch:
                   While (p^.Rgt <> NIL)
                     and (User^.Compare(ItemKey, User^.KeyOf(p^.Rgt^.Itm)) = 0) Do
                     Begin
                        Inc(Index, Succ(p^.Idx and $00FFFFFF));
                        p := p^.Rgt;
                     End;
             End{Case};
             Inc(Index, Succ(p^.Idx and $00FFFFFF));
             Exit;
          End;
       End;
    End;
    Search := FALSE;
END;


FUNCTION tAVLCollection.FirstThat(Action: Pointer): Boolean; Assembler;
{ =================================================================== }
ASM
        push  DS
        lds   SI, [ Self ]
        xor   DX, DX
        xor   AX, AX
        mov   [ SI.Index.Word.0 ], AX
        mov   [ SI.Index.Word.2 ], AX
        mov   [ SI.Item.Word.0 ], AX
        mov   [ SI.Item.Word.2 ], AX
        lds   SI, [ SI.Root ]
        mov   CX, OFFSET @Done
        push  CX

@Next:  mov   CX, DS
        jcxz  @Rtn

        push  DS
        push  SI
        lds   SI, [ tAVLnode(SI).Lft ]
        call  @Next

        pop   SI; pop DS; push DS; push SI { Recover node address...}
        les   DI, [ tAVLnode(SI).Itm ]
        lds   SI, [ Self ]
        add   AX,1                      { We maintain local index to  }
        adc   DX,0                      { protect ourselves from user }
        mov   [ SI.Index.Word.0 ], AX   { changing Collection's index }
        mov   [ SI.Index.Word.2 ], DX
        mov   [ SI.Item.Word.0 ], DI
        mov   [ SI.Item.Word.2 ], ES
        push  DX
        push  AX
        push  ES
        push  DI
        push  [ BP.Word.0 ]
        mov   DS, [ BP-2 ]
        call  DWORD PTR [ Action ]
        or    AL, AL
        jnz   @Okay

        pop   AX
        pop   DX
        pop   SI
        pop   DS
        lds   SI, [ tAVLnode(SI).Rgt ]
        jmp   @Next
@Rtn:   retn

@Done:  xor   AL, AL
@Okay:  mov   DS, [ BP-2 ]
        mov   SP, BP
END;


PROCEDURE tAVLCollection.Traverse(Fwd: Boolean; Action: Pointer); Assembler;
{ =================================================================== }
ASM
        push  DS
        lds   SI, [ Self ]
        xor   AX, AX
        mov   [ SI.Item.Word.0 ], AX
        mov   [ SI.Item.Word.2 ], AX
        mov   DX, [ SI.Count.Word.0 ]
        mov   [ SI.Index.Word.0 ], DX
        mov   DX, [ SI.Count.Word.2 ]
        mov   [ SI.Index.Word.2 ], DX
        lds   SI, [ SI.Root ]
        mov   DX, OFFSET @Done
        push  DX

@Next:  mov   CX, DS
        jcxz  @Rtn

        inc   AX
        push  DS
        push  SI
        lds   SI, [ tAVLnode(SI).Rgt ]
        call  @Next

        pop   SI; pop DS;                       { Recover node address...}
        push  DS; push SI; push AX
        push  DS; push SI; push AX
        les   DI, [ tAVLnode(SI).Itm ]
        lds   SI, [ Self ]
        mov   [ SI.Item.Word.0 ], DI
        mov   [ SI.Item.Word.2 ], ES
        mov   DS, [ BP-2 ]
        push  [ BP.Word.0 ]
        call  DWORD PTR [ Action ]
        lds   SI, [ Self ]
        sub   [ SI.Index.Word.0 ], 1
        sbb   [ SI.Index.Word.2 ], 0
        pop   AX
        pop   SI
        pop   DS
        lds   SI, [ tAVLnode(SI).Lft ]
        call  @Next
        dec   AX
@Rtn:   retn

@Done:  mov   DS, [ BP-2 ]
        mov   SP, BP
END;


FUNCTION  tAVLCollection.LastThat(Action: Pointer): Boolean; Assembler;
{ =================================================================== }
ASM
        push  DS
        lds   SI, [ Self ]
        xor   AX, AX
        xor   DX, DX
        mov   [ SI.Item.Word.0 ], AX
        mov   [ SI.Item.Word.2 ], AX
        mov   [ SI.Index.Word.0 ],AX
        mov   [ SI.Index.Word.0 ],AX

        lds   SI, [ SI.Root ]
        mov   BX, OFFSET @Count
        mov   CX, OFFSET @Done
        push  CX

@Next:  mov   CX, DS
        jcxz  @Rtn

        push  DS
        push  SI
        jmp   BX

@Count: stc
        adc   AX,[ tAVLnode(SI).Idx.Word.0 ]
        adc   DL,[ tAVLnode(SI).Idx.Byte.2 ]
        adc   DH,0

@OnRGT: lds   SI, [ tAVLnode(SI).Rgt ]
        call  @Next

        pop   SI; pop DS; push  DS; push SI  { Refresh node address }
        les   DI, [ tAVLnode(SI).Itm ]
        lds   SI, [ Self ]
        mov   [ SI.Index.Word.2 ], DX
        mov   [ SI.Index.Word.0 ], AX
        mov   [ SI.Item.Word.0 ], DI
        mov   [ SI.Item.Word.2 ], ES
        sub   AX, 1
        sbb   DX, 0
        push  DX
        push  AX
        push  ES
        push  DI
        push  [ BP.Word.0 ]
        mov   DS, [ BP-2 ]
        call  [ Action ]
        or    AL, AL
        jnz   @Okay

        pop   AX
        pop   DX
        pop   SI
        pop   DS

        lds   SI, [ tAVLnode(SI).Lft ]
        mov   BX, OFFSET @OnRGT
        jmp   @Next

@Rtn:   retn

@Done:  xor   AL, AL
@Okay:  mov   DS, [ BP-2 ]
        mov   SP, BP
END;


FUNCTION tAVLCollection.NextThat(Subscript: Longint; Action: Pointer): Boolean; Assembler;
{ =================================================================== }
ASM
            push  DS
            xor   AX, AX
            xor   BX, BX
            lds   SI, [ Self ]
            mov   [ Index.Word.2 ], AX               { Initialize to zero    }
            mov   [ Index.Word.0 ], AX               { for case 0 to MaxLong }
            mov   [ Item.Word.2 ], AX
            mov   [ Item.Word.0 ], AX
            lea   SI, [ DI.Root ]
            mov   CX, OFFSET @Done
            push  CX

            { --------------------------------------------- }
            { This is the common beginning of the recursion }
            { --------------------------------------------- }

@Push:      lds   SI, [ SI ]                    { De-reference }
            mov   CX, DS
            jcxz  @Done

            push  DS                            { Save current NODE }
            push  SI
            xor   DH, DH
            mov   DL, [ tAVLnode(SI).Idx.Byte.2 ]
            mov   CX, [ tAVLnode(SI).Idx.Word.0 ]
            stc
            adc   CX, AX
            adc   BX, DX

            cmp   [ Subscript.Word.2 ], BX
            jl    @AtLFT
            jg    @AtRGT

            cmp   [ Subscript.Word.0 ], CX
            jb    @AtLFT
            ja    @AtRGT

            { DS:SI is the indexed node and DX:CX conatins the index }
            { change index to DX:BX and continue on right...         }
            mov   BX, CX

@OnRGT:     pop   SI
            pop   DS

            lds   SI, [ tAVLnode(SI).Rgt ]
@Next:      mov   CX, DS
            jcxz  @Rtn

            push  DS
            push  SI
            lds   SI, [ SI ]
            call  @Next

@Here:      pop   SI; pop DS; push DS; push SI { Recover node address...}
            les   DI, [ tAVLnode(SI).Itm ]
            lds   SI, [ Self ]
            add   BX,1
            adc   DX,0
            mov   [ SI.Index.Word.0 ], BX
            mov   [ SI.Index.Word.2 ], DX
            mov   [ SI.Item.Word.0 ], DI
            mov   [ SI.Item.Word.2 ], ES
            mov   DS, [ BP-2 ]
            push  DX
            push  BX
            push  ES
            push  DI
            push  [ BP.Word.0 ]
            call  [ Action ]
            pop   BX
            pop   DX
            or    AL, AL
            jz    @OnRGT
            jmp   @Okay

{ ------------------------------------------------------------- }
@AtLFT:     call  @Push
            jmp   @Here            { process here and greater ! }

@AtRGT:     mov   AX, DX
            mov   BX, CX
            lea   SI, [ tAVLnode(SI).Rgt ]
            call  @Push            { Below start...             }
            add   SP, 4
@Rtn:       retn

@Done:      xor   AL, AL
@Okay:      mov   DS, [ BP-2 ]
            mov   SP, BP
END;


FUNCTION tAVLCollection.PrevThat(Subscript: Longint; Action: Pointer): Boolean; Assembler;
{ ------------------------------------------------------------- }
ASM
            push  DS
            xor   AX, AX
            xor   BX, BX
            lds   SI, [ Self ]
            mov   [ SI.Index.Word.2 ], AX           { for case MaxLong to 0 }
            mov   [ SI.Index.Word.0 ], AX
            mov   [ Item.Word.2 ], AX
            mov   [ Item.Word.0 ], AX
            lea   SI, [ SI.Root ]
            mov   CX, OFFSET @Done
            push  CX

            { --------------------------------------------- }
            { This is the common beginning of the recursion }
            { --------------------------------------------- }
@Push:      lds   SI, [ SI ]
            mov   CX, DS
            jcxz  @Done

            push  DS                            { Save current NODE }
            push  SI
            xor   DH, DH
            mov   DL, [ tAVLnode(SI).Idx.Byte.2 ]
            mov   CX, [ tAVLnode(SI).Idx.Word.0 ]
            stc
            adc   CX, BX
            adc   DX, AX

            cmp   [ Subscript.Word.2 ], DX
            jl    @AtLFT
            jg    @AtRGT

            cmp   [ Subscript.Word.0 ], CX
            jb    @AtLFT
            ja    @AtRGT

            { DS:SI is the indexed node and DX:CX conatins the index }
            { change index to DX:BX and continue on left...          }
            mov   BX, CX

@OnLFT:     pop   SI
            pop   DS

@Next:      lds   SI, [ SI ]
            mov   CX, DS
            jcxz  @Rtn

            push  DS
            push  SI

@OnRGT:     lea   SI, [ tAVLnode(SI).Rgt ]
            call  @Next

@Here:      pop   SI; pop DS; push DS; push SI  { Recover node address...}
            les   DI, [ tAVLnode(SI).Itm ]
            lds   SI, [ Self ]
            sub   BX, 1
            sbb   DX, 0
            mov   [ SI.Index.Word.0 ], BX
            mov   [ SI.Index.Word.2 ], DX
            mov   [ SI.Item.Word.0 ], DI
            mov   [ SI.Item.Word.2 ], ES
            mov   DS, [ BP-2 ]
            push  DX
            push  BX
            push  ES
            push  DI
            push  [ BP.Word.0 ]
            call  [ Action ]
            pop   BX
            pop   DX
            or    AL, AL
            jz    @OnLFT
            jmp   @Okay

{ ------------------------------------------------------------- }
@AtLFT:     call  @Push
            add   SP, 4
@Rtn:       retn

@AtRGT:     mov   BX, CX
            mov   AX, DX
            lea   SI, [ tAVLnode(SI).Rgt ]
            call  @Push                         { Above start... }
            jmp   @Here

@Done:      xor   AL, AL
@Okay:      mov   DS, [ BP-2 ]
            mov   SP, BP
END;


{ =================================================================== }
{               A V L     I N S E R T     M E T H O D S               }
{ =================================================================== }

{$IFDEF AVL_DEBUG}
PROCEDURE Trace(i: Word);
{ ------------------------------------------------------------------- }
TYPE Id =  Array[1..2] Of Char;
VAR   j: Integer;
BEGIN
     j := Swap(i);
     j := 2 * (Binary(Id(j))+3);
     If Adjust[j] < '2' Then
        Inc(Adjust[j]);
     Inc(Tally[(j div 2) -3]);
     Rotate := Char(Hi(i)) + Char(Lo(i));
END;
{$ENDIF}


FUNCTION AddedLeft(VAR p: tpAVLnode; Adj: Boolean): Boolean;  NEAR; Assembler;
{ ------------------------------------------------------------------- }
{ NOTE: Called with AL & Z-Flag reflecting need for adjustment ...    }
{ ------------------------------------------------------------------- }
CONST       DSreg =  -2;
            p1    =  -6;
            p2    = -10;
ASM
            push  DS
            lds   SI,[ p ]                      { Retrieve VAR parameter        }
            lds   SI,[ SI ]                     { De-reference                  }
            mov   AX, 1
            add   [tAVLnode(SI).Idx.Word.0],AX  { Increase number of left nodes }
            adc   [tAVLnode(SI).Idx.Byte.2],AH
            and   AL, [ Adj ]
            jz    @Exit

            cmp   [tAVLnode(SI).Bal],AL
            jz    @ClrBal                       { Case BAL = +1, Previous Right, Now balanced  }
            jc    @NegBal                       { Case BAL =  0, Previous balance, now heavy left }
                                                { Case BAL = -1, Previous Left, Now unbalanced Left }
            { -------------------------- }
            { p1 := p^.Lft               }
@@11:       les   DI,[tAVLnode(SI).Lft]

            { -------------------------- }
            { If p1^.Bal = -1            }
            add   AL, [tAVLnode(ES:DI).Bal]
            jnc   @@14

            { -------------------------- }
            { Single Right Rotation      }
            { -------------------------- }

            { -------------------------- }
            { Dec(p^.Idx,p1^.Idx+1);     }

            { Carry set from test for Bal = -1 ... }
            mov   AX, [tAVLnode(ES:DI).Idx.Word.0]
            sbb   [tAVLnode(DS:SI).Idx.Word.0], AX
            mov   AL, [tAVLnode(ES:DI).Idx.Byte.2]
            sbb   [tAVLnode(DS:SI).Idx.Byte.2], AL

            { -------------------------- }
            { p^.Lft := p1^.Rgt          }
            { p1^.Rgt := p;              }

            mov   AX,DS                              { msw of p }
            xchg  [tAVLnode(ES:DI).Rgt.Word.2],AX    { exchanged with msw of p1^.Rgt }
            mov   [tAVLnode(DS:SI).Lft.Word.2],AX    { and stored to msw of p^.Lft }

            mov   AX,SI                              { lsw of p }
            xchg  [tAVLnode(ES:DI).Rgt.Word.0],AX    { exchanged with lsw of p1^.Rgt }
            mov   [tAVLnode(DS:SI).Lft.Word.0],AX    { and stored to lsw of p^.Lft }

            { -------------------------- }
            { p^.Bal := 0;               }
            mov   [tAVLnode(DS:SI).Bal],0

            { -------------------------- }
            { p := p1                    }
            lds    SI,[ p ]
            mov    [DS:SI.Word.2],ES
            mov    [DS:SI.Word.0],DI
{$IFDEF AVL_DEBUG}
            mov    AX, '-1';
            push   AX
            mov    DS,[BP.DSreg]
            call   Trace
{$ENDIF}
            jmp    @EndAdj

            { -------------------------- }
            { Double left-right rotation }
            { -------------------------- }

            { -------------------------- }
            { Save p1                    }
@@14:       push   ES
            push   DI

            { -------------------------- }
            { p2 := p1^.Rgt              }
            lds    SI,[tAVLnode(ES:DI).Rgt]
            push   DS
            push   SI

            { ------------------------------ }
            { Dec(p^.Idx,p1^.Idx+p2^.Idx+2); }
            { Inc(p2^.Idx,p1^.Idx+1);        }

            mov    AX,[tAVLnode(ES:DI).Idx.Word.0]   { p1^.Idx }
            mov    DL,[tAVLnode(ES:DI).Idx.Byte.2]
            push   DX; push AX
            stc
            adc    AX,[tAVLnode(DS:SI).Idx.Word.0]   { + p2^.Idx }
            adc    DL,[tAVLnode(DS:SI).Idx.Byte.2]
            lds    SI, [ p ]
            lds    SI, [ SI ]
            stc
            sbb    [tAVLnode(DS:SI).Idx.Word.0],AX
            sbb    [tAVLnode(DS:SI).Idx.Byte.2],DL

            pop    AX; pop DX
            lds    SI, [ BP.p2 ]
            stc
            adc    [tAVLnode(DS:SI).Idx.Word.0],AX
            adc    [tAVLnode(DS:SI).Idx.Byte.2],DL

            { -------------------------- }
            { p1^.Rgt := p2^.Lft         }
            { p2^.Lft := p1              }

            mov   AX,ES                              { msw of p1 }
            xchg  [tAVLnode(DS:SI).Lft.Word.2],AX    { exchanged with msw of p2^.Lft }
            mov   [tAVLnode(ES:DI).Rgt.Word.2],AX    { and stored to msw of p1^.Rgt }

            mov   AX,DI                              { lsw of p1 }
            xchg  [tAVLnode(DS:SI).Lft.Word.0],AX    { exchanged with lsw of p2^.Lft }
            mov   [tAVLnode(ES:DI).Rgt.Word.0],AX    { and stored to lsw of p1^.Rgt }

            { -------------------------- }
            { p^.Lft := p2^.Rgt          }
            { p2^.Rgt := p               }

            les  DI,[ p ]                            { Retrieve VAR p }
            les  DI,[ES:DI]                          { and De-reference }

            mov   AX,ES                              { msw of p }
            xchg  [tAVLnode(DS:SI).Rgt.Word.2],AX    { exchanged with msw of p2^.Rgt }
            mov   [tAVLnode(ES:DI).Lft.Word.2],AX    { and stored to msw of p^.Lft }

            mov   AX,DI                              { lsw of p }
            xchg  [tAVLnode(DS:SI).Rgt.Word.0],AX    { exchanged with lsw of p2^.Rgt }
            mov   [tAVLnode(ES:DI).Lft.Word.0],AX    { and stored to lsw of p^.Lft }

            { --------------------------- }
            { p^.Bal := Ord(p2^.Bal = -1) }

            mov   AL,[tAVLnode(DS:SI).Bal]           { Retrieve p2^.Bal }
            xor   CX,CX                              { clear double byte }
            cmp   AL,0FFh
            jnz   @@15
            inc   CH                                 { TRUE set CH to +1 }

@@15:       mov   [tAVLnode(ES:DI).Bal],CH           { Store result to p^.Bal }

            { ---------------------------- }
            { p1^.Bal := -Ord(p2^.Bal = 1) }
            cmp   AL,1
            jnz   @@16
            dec   CL                                 { TRUE set CL to -1 }

@@16:       lds   SI,[ BP.p1 ]                       { Retrieve p1 }
            mov   [tAVLnode(DS:SI).Bal],CL           { and store result to p1^.Bal }

            { p := p2 }
            lds   SI,[ p ]                           { Retrieve VAR p }
            pop   [DS:SI.Word.0]                     { and store p2 to p }
            pop   [DS:SI.Word.2]
{$IFDEF AVL_DEBUG}
            mov    AX, '-2';
            push   AX
            mov    DS,[BP.DSreg]
            call   Trace
{$ENDIF}

@EndAdj:    lds   SI, [ p ]
            lds   SI, [ DS:SI ]

@ClrBal:    xor   AL, AL
@NegBal:    neg   AL
            mov   [tAVLnode(DS:SI).Bal],AL
@Exit:      mov   DS, [ BP.DSreg ]
            mov   SP, BP

END {AddedLeft};


FUNCTION AddedRight(VAR p: tpAVLnode; Adj: Boolean): Boolean; NEAR; Assembler;
{ ------------------------------------------------------------------- }
{ NOTE: Called with AL & Z-Flag reflecting need for adjustment ...    }
{ ------------------------------------------------------------------- }
CONST       DSreg =  -2;
            p1    =  -6;
            p2    = -10;
ASM
            mov   AL, 1
            and   AL, [ Adj ]
            jz    @Exit

            push  DS
            lds   SI,[ p ]                           { Retrieve VAR parameter }
            lds   SI,[DS:SI]                         { De-reference }
            cmp   [tAVLnode(DS:SI).Bal],AL
            jc    @SetBal                            { Case BAL =  0, Previous balance, now heavy right    }
            jnz   @ClrBal                            { Case BAL = -1, Previous left, now balanced          }
                                                     { Case BAL = +1, Previous right, now unbalanced right }
            { -------------------------- }
            { p1 = p^.Rgt                }
            les   DI,[tAVLnode(DS:SI).Rgt]           { Case BAL =  1 }

            { -------------------------- }
            { If p1^.Bal = 1             }
            cmp  [tAVLnode(ES:DI).Bal],AL
            jnz  @@23

            { ---------------------------- }
            { perform single left rotation }
            { ---------------------------- }

            { -------------------------- }
            { Inc(p1^.Idx,p^.Idx+1);     }

            mov  AX,[tAVLnode(DS:SI).Idx.Word.0]
            mov  DL,[tAVLnode(DS:SI).Idx.Byte.2]
            stc
            adc  [tAVLnode(ES:DI).Idx.Word.0],AX
            adc  [tAVLnode(ES:DI).Idx.Byte.2],DL

            { -------------------------- }
            { p^.Rgt := p1^.Lft;         }
            { p1^.Lft := p;              }

            mov   AX,DS                              { msw of p }
            xchg  [tAVLnode(ES:DI).Lft.Word.2],AX    { exchanged with msw of p1^.Lft }
            mov   [tAVLnode(DS:SI).Rgt.Word.2],AX    { and stored to msw of p^.Rgt }

            mov   AX,SI                              { lsw of p }
            xchg  [tAVLnode(ES:DI).Lft.Word.0],AX    { exchanged with lsw of p1^.Lft }
            mov   [tAVLnode(DS:SI).Rgt.Word.0],AX    { and stored to lsw of p^.Rgt }

            { -------------------------- }
            { p^ .Bal := 0;              }
            mov  [tAVLnode(DS:SI).Bal],0

            { -------------------------- }
            { p := p1;                   }
            lds  SI,[ p ]
            mov  [DS:SI.Word.2],ES
            mov  [DS:SI.Word.0],DI
{$IFDEF AVL_DEBUG}
            mov    AX, '+1';
            push   AX
            mov    DS,[BP.DSreg]
            call   Trace
{$ENDIF}
            jmp  @EndAdj

            { ---------------------------------- }
            { perform double right-left rotation }
            { ---------------------------------- }

            { -------------------------- }
            { Save p1                    }
@@23:       push  ES
            push  DI

            { -------------------------- }
            { p2 := p1^.Lft;             }
            mov   CX, [tAVLnode(DS:SI).Idx.Word.0]   { get  p^.Idx   }
            mov   BL, [tAVLnode(DS:SI).Idx.Byte.2]

            lds   SI,[tAVLnode(ES:DI).Lft]           { Init p2       }
            push  DS
            push  SI

            { -------------------------- }
            { Dec(p1^.Idx,Succ(p2^.Idx)) }
            { Inc(p2^.Idx,Succ(p^.Idx))  }
            mov  AX, [tAVLnode(DS:SI).Idx.Word.0]    { Get p2^.Idx   }
            mov  DL, [tAVLnode(DS:SI).Idx.Byte.2]
            stc
            sbb  [tAVLnode(ES:DI).Idx.Word.0],AX     { New p1^.Idx   }
            sbb  [tAVLnode(ES:DI).Idx.Byte.2],DL

            stc
            adc  [tAVLnode(DS:SI).Idx.Word.0],CX     { New p2^.Idx   }
            adc  [tAVLnode(DS:SI).Idx.Byte.2],BL

            { -------------------------- }
            { p1^.Lft := p2^.Rgt;        }
            { p2^.Rgt := p1;             }

            mov   AX,ES                              { msw of p1 }
            xchg  [tAVLnode(DS:SI).Rgt.Word.2],AX    { msw p2^.Rgt := p1      }
            mov   [tAVLnode(ES:DI).Lft.Word.2],AX    { msw p1^.Lft := p2^.Rgt }

            mov   AX,DI                              { lsw of p1 }
            xchg  [tAVLnode(DS:SI).Rgt.Word.0],AX    { lsw p2^.Rgt := p1      }
            mov   [tAVLnode(ES:DI).Lft.Word.0],AX    { lsw p1^.Lft := p2^.Rgt }

            { -------------------------- }
            { p^ .Rgt := p2^.Lft;        }
            { p2^.Lft := p;              }

            les   DI,[ p ]                           { Retrieve VAR p }
            les   DI,[ES:DI]                         { and dereference }

            mov   AX,ES                              { msw of p }
            xchg  [tAVLnode(DS:SI).Lft.Word.2],AX    { exchanged with msw of p2^.Lft }
            mov   [tAVLnode(ES:DI).Rgt.Word.2],AX    { and stored to msw of p^.Rgt }

            mov   AX,DI                              { lsw of p }
            xchg  [tAVLnode(DS:SI).Lft.Word.0],AX    { exchanged with lsw of p2^.Lft }
            mov   [tAVLnode(ES:DI).Rgt.Word.0],AX    { and stored to lsw of p^.Rgt }

            { ------------------------------ }
            { p^ .Bal := -Ord(p2^.Bal =  1); }

            mov  AL,[tAVLnode(DS:SI).Bal]            { retrieve p2^.Bal }
            xor  CX,CX                               { clear double byte }
            cmp  AL,1
            jnz  @@25
            dec  CH                                  { TRUE set CH to -1 }

@@25:       mov  [tAVLnode(ES:DI).Bal],CH            { Store result to p^.Bal }

            { ------------------------------ }
            { p1^.Bal :=  Ord(p2^.Bal = -1); }

            cmp  AL,0FFh
            jnz  @@26
            inc  CL                                  { TRUE Set CL to +1 }

@@26:       lds  SI,[ BP.p1 ]                        { Retrieve p1 }
            mov  [tAVLnode(DS:SI).Bal],CL            { and store reult to p1^.Bal }

            { p := p2; }
            lds  SI,[ p ]                            { Retrieve VAR p }
            pop  [DS:SI.Word.0]                      { and set p to p2 }
            pop  [DS:SI.Word.2]
{$IFDEF AVL_DEBUG}
            mov    AX, '+2';
            push   AX
            mov    DS,[BP.DSreg]
            call   Trace
{$ENDIF}

@EndAdj:    lds   SI, [ p ]
            lds   SI, [ DS:SI ]

@ClrBal:    xor   AL, AL
@SetBal:    mov   [tAVLnode(DS:SI).Bal],AL
            or    AL, AL
            mov   DS, [ BP.DSreg ]
            mov   SP, BP
@Exit:
END;{AddedRight}


FUNCTION tAVLCollection.Insert(TheItem: Pointer): Boolean; Assembler;
{ =================================================================== }
CONST      Subscript =  -4;   { Longint }
           Dupes     =  -6;   { Boolean }
           Strategy  = -10;   { Pointer }
           KeyOf     = -14;   { Pointer }
           Compare   = -18;   { Pointer }
           ItemKey   = -22;   { Pointer }
ASM
            les   DI, [ Self ]
            { >>> add code to watch for index overflow <<< }
            { >>> by examining index of root node ...  <<< }
            xor   AX, AX
            push  AX
            inc   AX
            push  AX
            inc   AL
            and   AL, [ tAVLCollection(ES:DI).Options.Byte.0 ]   { Duplicate flag }
            push  AX
            les   DI, [ tAVLCollection(ES:DI).User ]
            push  ES
            push  DI
            mov   BX, [ ES:DI ]
            push  [ BX.KeyOf_Method.Word.2 ]
            push  [ BX.KeyOf_Method.Word.0 ]
            push  [ BX.Compare_Method.Word.2 ]
            push  [ BX.Compare_Method.Word.0 ]
            push  [ TheItem.Word.2 ]
            push  [ TheItem.Word.0 ]
            push  ES                                 { user's self }
            push  DI
            call  DWORD PTR [ BP.KeyOf ]
            push  DX
            push  AX
            les   DI, [ Self ]
            lea   DI, [ tAVLCollection(DI).Root ]
            mov   AX, OFFSET @Done
            push  AX

@INSnode:   { ----------------------------------------------------- }
            { This is the common beginning of the recursion         }
            { VAR parameter being pushed is owned by INSlft/INSrgt  }
            { or by FreeItem if insert aborts               ...red  }
            { ----------------------------------------------------- }
            push  ES                                 { Save current (VAR p: NODE) }
            push  DI
            les   DI, [ ES:DI ]                      { Dereference       }
            mov   CX, ES                             { and check for nil }
            jcxz  @INShere

            push  ES                                 {  Save a copy ...           }
            push  DI
            push  [ BP.ItemKey.Word.2 ]
            push  [ BP.ItemKey.Word.0 ]
            push  [ tAVLnode(ES:DI).Itm.Word.2 ]
            push  [ tAVLnode(ES:DI).Itm.Word.0 ]
            push  [ BP.Strategy.Word.2 ]
            push  [ BP.Strategy.Word.0 ]
            call  DWORD PTR [ BP.Keyof ]
            push  DX
            push  AX
            push  [ BP.Strategy.Word.2 ]
            push  [ BP.Strategy.Word.0 ]
            call  DWORD PTR [ BP.Compare ]
            or    AL, [ BP.Dupes ]
            pop   DI                                 { Retrieve dereferenced NODE }
            pop   ES
            js    @InsLFT
            jz    @Fail

{ ------------------------------------------------------------- }
@InsRGT:    mov   AX, [ tAVLnode(ES:DI).Idx.Word.0 ]
            mov   DL, [ tAVLnode(ES:DI).Idx.Byte.2 ]
            xor   DH, DH
            stc
            adc   [ BP.Subscript.Word.0],AX
            adc   [ BP.Subscript.Word.2],DX
            lea   DI, [ tAVLnode(DI).Rgt ]
            call  @INSnode
            push  AX
            call  AddedRight
            retn

{ ------------------------------------------------------------- }
@InsLFT:    call  @INSnode
            push  AX
            call  AddedLeft
            retn

            { ------------------------------------------------- }
@INShere:   les   DI, [ Self ]
            add   [ ES:DI.Count.Word.0 ], 1          { Inc(Count) }
            adc   [ ES:DI.Count.Word.2 ], 0
            mov   AX, [ BP.Subscript.Word.0 ]        { Update Index }
            mov   [ ES:DI.Index.Word.0 ], AX
            mov   AX, [ BP.Subscript.Word.2 ]
            mov   [ ES:DI.Index.Word.2 ], AX
            mov   AX, [ TheItem.Word.2 ]             { Update Item   }
            mov   [ ES:DI.Item.Word.2 ], AX
            mov   AX, [ TheItem.Word.0 ]
            mov   [ ES:DI.Item.Word.0 ], AX
            push  [ TheItem.Word.2 ]
            push  [ TheItem.Word.0 ]
            push  ES
            push  DI
            call  BuildNode
            or    AL, 1                              { Set NZ to adjust balance ... }
            retn

            { -------------------------------------------------- }
@Fail:      push  [ TheItem.Word.2 ]
            push  [ TheItem.Word.0 ]
            push  [ tAVLnode(ES:DI).Itm.Word.2 ]     { Save current item }
            push  [ tAVLnode(ES:DI).Itm.Word.0 ]
            les   DI, [ Self ]

            pop   [ ES:DI.Item.Word.0 ]              { Mark duped? item  }
            pop   [ ES:DI.Item.Word.2 ]
            mov   AX, [ BP.Subscript.Word.2 ]
            mov   [ ES:DI.Index.Word.2 ], AX
            mov   AX, [ BP.Subscript.Word.0 ]
            mov   [ ES:DI.Index.Word.0 ], AX

            les   DI, [ ES:DI.User ]                 { Free failed item  }
            push  ES
            push  DI
            mov   BX, [ ES:DI ]
            call  DWORD PTR [ BX.FreeItem_Method ]
{$IFDEF AVL_DEBUG}
            mov   AX,'**'
            push  AX
            call  Trace
{$ENDIF}
            xor   AX, AX                             { FALSE: Insert failed }
            jmp   @Exit

@Done:      mov   AL, 1                              { TRUE:  Inserted okay }
@Exit:      mov   SP, BP                             { Discard recursion    }

END {AVL.Insert};


PROCEDURE tAVLCollection.AtInsert(Subscript: Longint; TheItem: Pointer); Assembler;
{ =================================================================== }
ASM
            push  DS
            lds   SI, [ Self ]
            { >>> add code to watch for index overflow <<< }
            { >>> by examining index of root node ...  <<< }
            lea   SI, [ SI.Root ]
            xor   BX,BX
            xor   DX,DX
            mov   AX, OFFSET @Done
            push  AX

            { ------------------------------------------------------- }
            { This is the common beginning of the recurrsion.         }
            { VAR parameter being pushed is owned by INSlft/INSrgt    }
            { or will be consummed by InsHere                  ...red }
            { ------------------------------------------------------- }
@INSnode:   push  DS                     { Save current (VAR p: NODE) }
            push  SI
            lds   SI, [ SI ]             { Dereference       }
            mov   CX, DS                 { and check for nil }
            jcxz  @INShere

            mov   CX, [ tAVLnode(SI).Idx.Word.0 ]
            mov   AL, [ tAVLnode(SI).Idx.Byte.2 ]
            xor   AH, AH
            stc
            adc   CX, BX
            adc   AX, DX

            cmp   [ Subscript.Word.2 ], AX
            jg    @InsRGT
            jl    @InsLft

            cmp   [ Subscript.Word.0 ], CX
            jna   @InsLFT

{ ------------------------------------------------------------- }
@InsRGT:    mov   DX, AX
            mov   BX, CX
            lea   SI, [ tAVLnode(SI).Rgt ]
            call  @INSnode
            push  AX
            call  AddedRight
            retn

            { --------------------------------------------- }
@INShere:   lds   SI, [ Self ]
            stc
            adc   [ SI.Count.Word.0 ], CX            { Inc(Count)     }
            adc   [ SI.Count.Word.2 ], CX
            stc
            adc   BX, CX                             { Inc(Subscript) }
            adc   DX, CX
            mov   [ SI.Index.Word.0 ], BX            { Update Index   }
            mov   [ SI.Index.Word.2 ], DX
            mov   AX, [ TheItem.Word.2 ]             { Update Item    }
            mov   [ SI.Item.Word.2 ], AX
            push  AX
            mov   AX, [ TheItem.Word.0 ]
            mov   [ SI.Item.Word.0 ], AX
            push  AX
            push  DS
            push  SI
            mov   DS, [ BP-2 ]
            call  BuildNode
            or    AL, 1                { Set TRUE to adjust balance ... }
            retn

{ --------------------------------------------------- }
@InsLFT:    call  @INSnode
            push  AX
            call  AddedLeft
            retn

@Done:      mov   DS, [ BP-2 ]
            mov   SP, BP                          { Discard recursion }

END;{AVL.AtInsert}


END.
